#-----------------------------------------------------------------------
# File    : pycoco.mak
# Contents: build pycoco dynamic link library (on Windows systems)
# Author  : Christian Borgelt
# History : 2012.08.03 file created
#           2013.10.19 module patspec added
#           2013.11.08 properly adapted to Windows/Microsoft Visual C
#           2015.03.04 interrupt signal handling added
#           2015.03.05 preprocessor definitions XXX_ABORT added 
#           2015.08.19 module patred added
#           2016.04.20 completed dependencies on header files
#           2022.11.25 compilation for Python 2 removed
#-----------------------------------------------------------------------
!IFNDEF PYDIR
PYDIR    = "C:\Program Files\Python310"
!ENDIF
THISDIR  = ..\..\pycoco\src
UTILDIR  = ..\..\util\src
MATHDIR  = ..\..\math\src
TRACTDIR = ..\..\tract\src
COCODIR  = ..\..\coconad\src

CC       = cl.exe
DEFS     = /D WIN32 /D NDEBUG /D _CONSOLE /D _CRT_SECURE_NO_WARNINGS \
           /D QUIET
CFLAGS   = /nologo /MD /W3 /O2 /GS- $(DEFS) /c
PYINC    = /I $(PYDIR)\include
INCS     = /I $(UTILDIR) /I $(MATHDIR) /I $(TRACTDIR) /I $(COCODIR)

LD       = link.exe
LDFLAGS  = /DLL /nologo /incremental:no
LIBS     = /LIBPATH:$(PYDIR)\libs

HDRS     = $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h    \
           $(UTILDIR)\memsys.h    $(UTILDIR)\symtab.h    \
           $(UTILDIR)\sigint.h    $(TRACTDIR)\tract.h    \
           $(TRACTDIR)\train.h    $(TRACTDIR)\patspec.h  \
           $(TRACTDIR)\clomax.h   $(TRACTDIR)\cm4seqs.h  \
           $(TRACTDIR)\cmfilter.h $(TRACTDIR)\report.h
OBJS     = arrays.obj memsys.obj idmap.obj random.obj      \
           sigint.obj gamma.obj tract.obj train.obj        \
           patspec.obj clomax.obj cm4seqs.obj cmfilter.obj \
           report.obj patred.obj coconad.obj ccnpsp.obj    \
           pycoco.obj $(ADDOBJS)

OVLHDRS  = $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h    \
           $(UTILDIR)\memsys.h    $(UTILDIR)\symtab.h    \
           $(UTILDIR)\sigint.h    $(TRACTDIR)\tract.h    \
           $(TRACTDIR)\train.h    $(TRACTDIR)\patspec.h  \
           $(TRACTDIR)\clomax.h   $(TRACTDIR)\cmfilter.h \
           $(TRACTDIR)\report.h
OVLOBJS  = arrays.obj memsys.obj idmap.obj random.obj    \
           sigint.obj gamma.obj tads.obj trnds.obj       \
           pspdbl.obj cmdbl.obj cmfdbl.obj               \
           repcmd.obj predd.obj ccn-ovl.obj ccnpsp.obj   \
           pyc-ovl.obj $(ADDOBJS)
		   
#-----------------------------------------------------------------------
# Build Dynamic Link Library
#-----------------------------------------------------------------------
all:          coco.pyd ccnovl.pyd

coco.pyd:     $(OBJS) pycoco.mak
	$(LD) $(LDFLAGS) $(LIBS) $(OBJS) /out:$@ /IMPLIB:coco.lib

ccnovl.pyd:   $(OVLOBJS) pycoco.mak
	$(LD) $(LDFLAGS) $(LIBS) $(OVLOBJS) /out:$@ /IMPLIB:ccnovl.lib

#-----------------------------------------------------------------------
# Array Operations
#-----------------------------------------------------------------------
arrays.obj:   $(UTILDIR)\fntypes.h
arrays.obj:   $(UTILDIR)\arrays.h    $(UTILDIR)\arrays.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)\arrays.c /Fo$@

#-----------------------------------------------------------------------
# Memory Management System for Objects of Equal Size
#-----------------------------------------------------------------------
memsys.obj:   $(UTILDIR)\memsys.h    $(UTILDIR)\memsys.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)\memsys.c /Fo$@

#-----------------------------------------------------------------------
# Symbol Table Management
#-----------------------------------------------------------------------
idmap.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h
idmap.obj:    $(UTILDIR)\symtab.h    $(UTILDIR)\symtab.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D IDMAPFN $(UTILDIR)\symtab.c /Fo$@

#-----------------------------------------------------------------------
# Random Number Generator Management
#-----------------------------------------------------------------------
random.obj:   $(UTILDIR)\random.h    $(UTILDIR)\random.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)\random.c /Fo$@

#-----------------------------------------------------------------------
# Interrupt Signal Handling
#-----------------------------------------------------------------------
sigint.obj:   $(UTILDIR)\sigint.h    $(UTILDIR)\sigint.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)\sigint.c /Fo$@

#-----------------------------------------------------------------------
# Gamma Function
#-----------------------------------------------------------------------
gamma.obj:    $(MATHDIR)\gamma.h     $(MATHDIR)\gamma.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(MATHDIR)\gamma.c /Fo$@

#-----------------------------------------------------------------------
# Item and Transaction Management
#-----------------------------------------------------------------------
tract.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h
tract.obj:    $(TRACTDIR)\tract.h    $(TRACTDIR)\tract.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)\tract.c /Fo$@

tads.obj:     $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h
tads.obj:     $(TRACTDIR)\tract.h    $(TRACTDIR)\tract.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D SUPP=double $(TRACTDIR)\tract.c /Fo$@

#-----------------------------------------------------------------------
# Train/Point Process Management
#-----------------------------------------------------------------------
train.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h
train.obj:    $(TRACTDIR)\train.h    $(TRACTDIR)\train.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D TRN_SURR $(TRACTDIR)\train.c /Fo$@

trnds.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h
trnds.obj:    $(TRACTDIR)\train.h $(TRACTDIR)\train.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D TRN_SURR /D SUPP=double \
              $(TRACTDIR)\train.c /Fo$@

#-----------------------------------------------------------------------
# Item Set Reporter Management
#-----------------------------------------------------------------------
patspec.obj:  $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h   \
              $(UTILDIR)\symtab.h    $(UTILDIR)\tabwrite.h \
              $(TRACTDIR)\tract.h
patspec.obj:  $(TRACTDIR)\patspec.h  $(TRACTDIR)\patspec.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D PSP_ESTIM /D PSP_TRAIN \
	      $(TRACTDIR)\patspec.c /Fo$@

pspdbl.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h   \
              $(UTILDIR)\symtab.h    $(UTILDIR)\tabwrite.h \
              $(TRACTDIR)\tract.h
pspdbl.obj:   $(TRACTDIR)\patspec.h  $(TRACTDIR)\patspec.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D PSP_ESTIM /D PSP_TRAIN \
              /D SUPP=double $(TRACTDIR)\patspec.c /Fo$@

clomax.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\memsys.h    $(UTILDIR)\symtab.h \
              $(TRACTDIR)\tract.h
clomax.obj:   $(TRACTDIR)\clomax.h   $(TRACTDIR)\clomax.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)\clomax.c /Fo$@

cmdbl.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\memsys.h    $(UTILDIR)\symtab.h \
              $(TRACTDIR)\tract.h
cmdbl.obj:    $(TRACTDIR)\clomax.h   $(TRACTDIR)\clomax.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D RSUPP=double \
              $(TRACTDIR)\clomax.c /Fo$@

report.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h   \
              $(UTILDIR)\symtab.h    $(UTILDIR)\scanner.h  \
              $(TRACTDIR)\tract.h    $(TRACTDIR)\patspec.h \
              $(TRACTDIR)\clomax.h
report.obj:   $(TRACTDIR)\report.h   $(TRACTDIR)\report.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D ISR_PATSPEC /D ISR_CLOMAX \
              /D ISR_NONAMES $(TRACTDIR)\report.c /Fo$@

repcmd.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h   \
              $(UTILDIR)\symtab.h    $(UTILDIR)\scanner.h  \
              $(TRACTDIR)\tract.h    $(TRACTDIR)\patspec.h \
              $(TRACTDIR)\clomax.h
repcmd.obj:   $(TRACTDIR)\report.h   $(TRACTDIR)\report.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D ISR_PATSPEC /D ISR_CLOMAX \
              /D ISR_NONAMES /D RSUPP=double $(TRACTDIR)\report.c /Fo$@

#-----------------------------------------------------------------------
# Prefix Tree Management for Closed and Maximal Sequences
#-----------------------------------------------------------------------
cm4seqs.obj:  $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\memsys.h    $(UTILDIR)\symtab.h \
              $(TRACTDIR)\tract.h    $(TRACTDIR)\report.h
cm4seqs.obj:  $(TRACTDIR)\cm4seqs.h  $(TRACTDIR)\cm4seqs.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)\cm4seqs.c /Fo$@

#-----------------------------------------------------------------------
# List/Array Closed/Maximal Filter Management
#-----------------------------------------------------------------------
cmfilter.obj: $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h \
              $(TRACTDIR)\report.h
cmfilter.obj: $(TRACTDIR)\cmfilter.h $(TRACTDIR)\cmfilter.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)\cmfilter.c /Fo$@

cmfdbl.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h \
              $(TRACTDIR)\report.h
cmfdbl.obj:   $(TRACTDIR)\cmfilter.h $(TRACTDIR)\cmfilter.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D RSUPP=double \
              $(TRACTDIR)\cmfilter.c /Fo$@

#-----------------------------------------------------------------------
# Pattern Set Reduction Functions
#-----------------------------------------------------------------------
patred.obj:   $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h \
              $(TRACTDIR)\report.h
patred.obj:   $(TRACTDIR)\patred.h   $(TRACTDIR)\patred.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)\patred.c /Fo$@

predd.obj:    $(UTILDIR)\fntypes.h   $(UTILDIR)\arrays.h \
              $(UTILDIR)\symtab.h    $(TRACTDIR)\tract.h \
              $(TRACTDIR)\report.h
predd.obj:    $(TRACTDIR)\patred.h   $(TRACTDIR)\patred.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D RSUPP=double \
              $(TRACTDIR)\patred.c /Fo$@

#-----------------------------------------------------------------------
# CoCoNAD
#-----------------------------------------------------------------------
coconad.obj:  $(HDRS)                $(TRACTDIR)\cm4seqs.h \
              $(TRACTDIR)\cmfilter.h
coconad.obj:  $(COCODIR)\coconad.h   $(COCODIR)\coconad.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D CCN_ABORT $(COCODIR)\coconad.c /Fo$@

ccn-ovl.obj:  $(HDRS)
ccn-ovl.obj:  $(COCODIR)\coconad.h   $(COCODIR)\coconad.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D CCN_ABORT /D CCN_OVERLAP \
              /D SUPP=double $(COCODIR)\coconad.c /Fo$@

#-----------------------------------------------------------------------
# Pattern Spectrum Functions
#-----------------------------------------------------------------------
ccnpsp.obj:   $(HDRS)                $(UTILDIR)\random.h \
              $(UTILDIR)\sigint.h    $(COCODIR)\coconad.h
ccnpsp.obj:   $(COCODIR)\ccnpsp.h    $(COCODIR)\ccnpsp.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) /D CCN_ABORT $(COCODIR)\ccnpsp.c /Fo$@

#-----------------------------------------------------------------------
# Python Stuff
#-----------------------------------------------------------------------
pycoco.obj:   $(HDRS)                $(UTILDIR)\random.h    \
              $(UTILDIR)\sigint.h    $(TRACTDIR)\patred.h   \
              $(TRACTDIR)\cm4seqs.h  $(TRACTDIR)\cmfilter.h \
              $(COCODIR)\coconad.h   $(COCODIR)\ccnpsp.h
pycoco.obj:   pycoco.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(PYINC) pycoco.c /Fo$@

pyc-ovl.obj:  $(HDRS)                $(UTILDIR)\random.h    \
              $(UTILDIR)\sigint.h    $(TRACTDIR)\patred.h   \
              $(TRACTDIR)\cmfilter.h $(COCODIR)\coconad.h   \
              $(COCODIR)\ccnpsp.h
pyc-ovl.obj:  pycoco.c pycoco.mak
	$(CC) $(CFLAGS) $(INCS) $(PYINC) \
              /D CCN_OVERLAP /D SUPP=double pycoco.c /Fo$@

#-----------------------------------------------------------------------
# Install
#-----------------------------------------------------------------------
install:
	-@copy *.pyd ..\..\..\bin

#-----------------------------------------------------------------------
# Clean up
#-----------------------------------------------------------------------
clean:
	-@erase /Q *~ *.pyd *.obj *.idb *.pch *.exp *.lib $(PRGS)
